# Built-in modules #
import getpass

# Internal modules #
from plumbing.cache import expiry_every
from plumbing.common import which

# Third party modules #
import sh

# Constants #
user = getpass.getuser()

################################################################################
class ExistingJobs(object):
    """Parses the output of `squeue` calls"""

    queued_command = ["--array", "--format='%.8A %.9P %.25j %.8u %.14a %.2t %.19S %.10L %.8Q %.4C %.16R %.12f %E'", "--sort=-p,i", "-t", "PENDING", "-u", user]
    running_command = ["-t", "r,cg", "--array", "--format='%.8A %.9P %.25j %.8u %.14a %.2t %.19S %.10L %.6D %.4C %R'", "-S", "e", "-u", user]

    queued_params = ['jobid','partition','name','user','account','state',
                     'start_time','time_left','priority','cpus','nodelist',
                     'features','dependency']
    running_params = ['jobid','partition','name','user','account','state',
                     'start_time','time_left','nodes','cpus','nodelist']

    def __iter__(self): return iter(self.status)
    def __getitem__(self, key):
        if isinstance(key, slice): return self.status[key]
        elif isinstance(key, int): return self.status[key]
        elif isinstance(key, basestring): return [s for s in self if s['name'] == key][0]
        else: raise TypeError("Invalid argument type.")
    def __contains__(self, key): return key in [s['name'] for s in self]

    @property
    @expiry_every(seconds=30)
    def status(self):
        # Parse #
        self.queued = [line for line in sh.squeue(self.queued_command) if not line.startswith("'   JOBID")]
        self.running = [line for line in sh.squeue(self.running_command) if not line.startswith("'   JOBID")]
        # Structure #
        self.queued = [dict(zip(self.queued_params, line.strip("'\n").split())) for line in self.queued if line]
        self.running = [dict(zip(self.running_params, line.strip("'\n").split())) for line in self.running if line]
        # Add info #
        for job in self.queued: job['type'] = 'queued'
        for job in self.running: job['type'] = 'running'
        # Return #
        return self.running + self.queued

    def expire(self):
        if hasattr(self.status, '__cache__'): del self.status.__cache__

    @property
    def names(self):
        return [job['name'] for job in self.status]

################################################################################
class ExistingProjects(object):
    """Parses the output of `projinfo` on Uppmax"""
    params = ['name', 'used', 'allowed']

    def __iter__(self): return iter(self.status)
    def __getitem__(self, key):
        if isinstance(key, slice): return self.status[key]
        elif isinstance(key, int): return self.status[key]
        elif isinstance(key, str): return [s for s in self if s['name'] == key][0]
        else: raise TypeError("Invalid argument type.")
    def __contains__(self, key): return key in [s['name'] for s in self]
    def __nonzero__(self): return bool(self.status)

    @property
    @expiry_every(seconds=3600)
    def status(self):
        if not which('projinfo', safe=True): return None
        self.output = sh.projinfo()
        self.lines = [line.strip("'\n") for line in self.output if line.startswith('b')]
        cast = lambda x: (str(x[0]), float(x[1]), float(x[2]))
        self.items = [cast(line.split()) for line in self.lines]
        self.result = [dict(zip(self.params, item)) for item in self.items]
        self.result.sort(key = lambda x: x['used'])
        return self.result

################################################################################
jobs = ExistingJobs()
projects = ExistingProjects()



# Built-in modules #
import subprocess, sys

# Constants #
module_names = {
    "biopython":    "Bio",
    "ipython":      "IPython",
    "scikit-learn": "sklearn",
    "scikit-bio":   "skbio",
    "biom-formt":   "biom",
}

################################################################################
def check_setup_py(mod_name):
    """Parses the required modules from a setup.py file and checks they are
    importable (e.g. available in the PYTHONPATH)"""
    pass

################################################################################
def check_executable(tool_name):
    """Raises an warning if the executable *tool_name* is not found."""
    result = subprocess.call(['which', tool_name], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    if result != 0:
        message = "The executable '%s' cannot be found in your $PATH" % tool_name
        raise Exception(message)

################################################################################
def check_module(mod_name):
    """Calls sys.exit() if the module *mod_name* is not found."""
    # Special cases #
    if mod_name in module_name: mod_name = module_name[mod_name]
    # Use a try except block #
    try:
        __import__(mod_name)
    except ImportError as e:
        if str(e) != 'No module named %s' % mod_name: raise e
        print 'You do not seem to have the "%s" package properly installed.' \
              ' Either you never installed it or your $PYTHONPATH is not set up correctly.' \
              ' For more instructions see the README file. (%s)' % (mod_name, e)
        sys.exit()


# Built-in modules #
import os, re, multiprocessing

# Internal modules #
from plumbing.common import is_integer

# Third party modules #

################################################################################
def count_processors():
    """How many cores does the current computer have ?"""
    if 'SLURM_NTASKS' in os.environ: return int(os.environ['SLURM_NTASKS'])
    elif 'SLURM_JOB_CPUS_PER_NODE' in os.environ:
         text = os.environ['SLURM_JOB_CPUS_PER_NODE']
         if is_integer(text): return int(text)
         else:
            n, N = re.findall("([1-9]+)\(x([1-9]+)\)", text)[0]
            return int(n) * int(N)
    else: return multiprocessing.cpu_count()

################################################################################
def guess_server_name():
    """We often use the same servers, which one are we running on now ?"""
    if   os.environ.get('CSCSERVICE')          == 'sisu':         return "sisu"
    elif os.environ.get('SLURM_JOB_PARTITION') == 'halvan':       return "halvan"
    elif os.environ.get('SNIC_RESOURCE')       == 'milou':        return "milou"
    elif os.environ.get('LAPTOP')              == 'macbook_air':  return "macbook_air"
    else:                                                         return "unknown"

################################################################################
num_processors = count_processors()
current_server = guess_server_name()


# Built-in modules #
import os, time
import dateutil.tz, datetime

# Internal modules #
from plumbing.slurm.job import JobSLURM
from plumbing.autopaths import DirectoryPath
from plumbing.git       import GitRepo

# Third party modules #
import sh

################################################################################
class LoggedJobSLURM(JobSLURM):
    """Takes care of running a python job through SLURM and logs results.
    Will run it remotely in a new interpreter with a static copy of all
    required modules."""

    def __init__(self, command,
                 language = 'python',
                 base_dir = os.path.abspath(os.getcwd()),
                 modules  = None,
                 **kwargs):
        # Check the modules variable is a list #
        if modules is None:                 self.modules = []
        elif not isinstance(modules, list): self.modules = list(modules)
        else:                               self.modules = modules
        # Check command type #
        if not isinstance(command, list): command = [command]
        # Log directory #
        for i in range(30):
            now = datetime.datetime.now(dateutil.tz.tzlocal())
            log_name = now.strftime("%Y-%m-%da%Hh%Mm%Ss%Z%z")
            base_dir = DirectoryPath(base_dir + log_name + '/')
            if not base_dir.exists:
                base_dir.create()
                break
            else:
                time.sleep(2)
                continue
        else: base_dir.create()
        # Modules directory #
        modules_dir = DirectoryPath(base_dir + "modules/")
        modules_dir.create()
        # The script to be sent #
        script =  []
        # Copy modules to the log directory #
        for module in self.modules:
            module_dir        = os.path.dirname(module.__file__)
            module_name       = module.__name__
            repos_dir         = GitRepo(os.path.abspath(module_dir + '/../'))
            project_name      = os.path.basename(repos_dir)
            static_module_dir = modules_dir + project_name + '/'
            module_version    = module.__version__ + ' ' + repos_dir.tag
            # Copy #
            print "Making static copy of module '%s' for SLURM job..." % module_name
            sh.cp('-R', repos_dir, static_module_dir)
            # Make script #
            script.insert(0, "sys.path.insert(0, '%s')" % static_module_dir)
            script += ["import %s" % module_name]
            script += ["print 'Using static copy of module %s version %s'" % (module_name, module_version)]
        # Prepend to the script to be sent #
        script.insert(0, "import os, sys")
        # Add the user's command to the script #
        script += command
        # Super #
        JobSLURM.__init__(self, script, language, base_dir, **kwargs)


# Built-in modules #
from datetime import datetime

# Internal modules #
from plumbing.color import Color

################################################################################
class Timer(object):
    """Useful for timing the different steps in a pipeline.

    Use it like this:

        timer = Timer()
        timer.print_start()
        timer.print_elapsed()
        timer.print_end()
        timer.print_total_elapsed()

    Or like this:

        with Timer():
            time.sleep(2)
    """

    def __init__(self):
        self.start_time = datetime.now()
        self.last_mark = datetime.now()

    def print_start(self):
        print self.prefix + "Start at: %s" % (self.start_time) + self.suffix

    def print_end(self):
        self.end_time = datetime.now()
        print self.prefix + "End at: %s" % (self.end_time) + self.suffix

    def print_elapsed(self, reset=True):
        print self.prefix + "Elapsed time: %s" % (datetime.now() - self.last_mark) + self.suffix
        if reset: self.last_mark = datetime.now()

    def print_total_elapsed(self, reset=True):
        print self.prefix + "Total elapsed time: %s" % (datetime.now() - self.start_time) + self.suffix
        if reset: self.last_mark = datetime.now()

    @property
    def color(self):
        """Should we use color or not ? If we are not in a shell like ipython then not."""
        import __main__ as main
        if not hasattr(main, '__file__'): return True
        return False

    @property
    def prefix(self): return "" if not self.color else Color.grn
    @property
    def suffix(self): return "" if not self.color else Color.end

    def __enter__(self):
        """Start the timer and print."""
        self.start_time = datetime.now()
        self.print_start()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """Stop the timer and print."""
        self.print_end()
        self.print_total_elapsed()


"""
=======================
Threads and parallelism
=======================

The ``thread`` module provides some convenience functions
for using threads in python.
"""

# Modules #
import threading

################################################################################
class ReturnThread(threading.Thread):
    """Like a normal thread, but the target function's return value is captured."""
    def __init__(self, group=None, target=None, name=None, args=(), kwargs={}, verbose=None):
        threading.Thread.__init__(self, group, target, name, args, kwargs, verbose)
        self._return = None

    def run(self):
        try:
            if self._Thread__target:
                self._return = self._Thread__target(*self._Thread__args, **self._Thread__kwargs)
        finally:
            del self._Thread__target, self._Thread__args, self._Thread__kwargs

    def join(self, timeout=None):
        threading.Thread.join(self, timeout)
        return self._return

################################################################################
def non_blocking(func):
    """Decorator to run a function in a different thread.
    It can be used to execute a command in a non-blocking way
    like this::

        @non_blocking
        def add_one(n):
            print 'starting'
            import time
            time.sleep(2)
            print 'ending'
            return n+1

        thread = add_one(5) # Starts the function
        result = thread.join() # Waits for it to complete
        print result
    """
    from functools import wraps
    @wraps(func)
    def non_blocking_version(*args, **kwargs):
        t = ReturnThread(target=func, args=args, kwargs=kwargs)
        t.start()
        return t
    return non_blocking_version



# Built-in modules #
import sys, time, datetime

# Internal modules #
from plumbing.common import iflatten
from plumbing.color import Color
from plumbing.slurm.logged import LoggedJobSLURM

# Third party modules #
import threadpool

# Constants #

###############################################################################
class Runner(object):
    """General purpose runner. Can execute functions on objects, via SLURM
    or locally. You should inherit from this class."""
    modules       = []
    job_name      = "unnamed"
    default_time  = '7-00:00:00'
    default_steps = [{'run': {}}]

    @property
    def extra_slurm_params(self): return {}

    def __init__(self, parent):
        self.parent = parent

    @property
    def color(self):
        """Should we use color or not ? If we are not in a shell, then not."""
        import __main__ as main
        if not hasattr(main, '__file__'): return True
        return False

    def run(self, steps=None, **kwargs):
        # Message #
        if self.color: print Color.f_cyn + "Running %s" % (self.parent) + Color.end
        else: print "Running %s" % self.parent
        # Check input #
        if not steps:                     steps = self.default_steps
        if isinstance(steps, basestring): steps = [steps]
        # Do all the steps #
        for step in steps:
            if isinstance(step, basestring):  name, params = step, {}
            if isinstance(step, dict):        name, params = step.items()[0]
            params.update(kwargs)
            fns = self.find_fns(name)
            self.run_step(name, fns, **params)
        # Report success #
        print "Success. Results are in %s" % self.parent.base_dir

    def find_fns(self, name):
        # Special case #
        if '.' in name:
            target = self.parent
            for attribute in name.split('.'):
                target = getattr(target, attribute)
            return [target]
        # Functions #
        def get_children(obj, name, level):
            if level == 0:
                if not hasattr(obj, name): return []
                else: return [getattr(obj, name)]
            else: return iflatten([get_children(o, name, level-1) for o in obj.children])
        # Recursive #
        fns = None
        level = 0
        while True:
            target = self.parent
            for i in range(level):
                if hasattr(target, 'first'): target = target.first
                else: target = None
            if target is None: raise Exception("Could not find function '%s'" % name)
            if hasattr(target, name):
                fns = get_children(self.parent, name, level)
                break
            level += 1
        assert fns
        return fns

    def run_step(self, name, fns, *args, **kwargs):
        # Default threads #
        if '.' in name: threads = kwargs.pop('threads', False)
        else:           threads = kwargs.pop('threads', True)
        # Start timer #
        start_time = time.time()
        # Message #
        if self.color: print "Running step: " + Color.f_grn + name + Color.end
        else: print "Running step: " + name
        sys.stdout.flush()
        # Threads #
        if threads and len(list(fns)) > 1:
            self.thpool = threadpool.ThreadPool(8)
            for fn in fns: self.thpool.putRequest(threadpool.WorkRequest(fn))
            self.thpool.wait()
            self.thpool.dismissWorkers(8)
            del self.thpool
        else:
            for fn in fns: fn(*args, **kwargs)
        # Stop timer #
        run_time = datetime.timedelta(seconds=round(time.time() - start_time))
        if self.color: print Color.ylw + "Run time: '%s'" % (run_time) + Color.end
        else: print "Run time: '%s'" % (run_time)
        sys.stdout.flush()

    @property
    def logs(self):
        """Find the log directory and return all the logs sorted."""
        if not self.parent.loaded: self.parent.load()
        logs = self.parent.p.logs_dir.flat_directories
        logs.sort(key=lambda x: x.mod_time)
        return logs

    @property
    def latest_log(self):
        """Find the latest log in all the logs."""
        return self.logs[-1]

    #-------------------------------------------------------------------------#
    def run_locally(self, steps=None, **kwargs):
        """A convenience method to run the same result as a SLURM job
        but locally in a non-blocking way."""
        self.slurm_job = LoggedJobSLURM(self.command(steps),
                                        base_dir = self.parent.p.logs_dir,
                                        modules  = self.modules,
                                        **kwargs)
        self.slurm_job.run_locally()

    #-------------------------------------------------------------------------#
    def run_slurm(self, steps=None, **kwargs):
        """Run the steps via the SLURM queue."""
        # Optional extra SLURM parameters #
        params = self.extra_slurm_params
        params.update(kwargs)
        # Mandatory extra SLURM parameters #
        if 'time'       not in params: params['time']       = self.default_time
        if 'job_name'   not in params: params['job_name']   = self.job_name
        if 'email'      not in params: params['email']      = None
        if 'dependency' not in params: params['dependency'] = 'singleton'
        # Send it #
        self.slurm_job = LoggedJobSLURM(self.command(steps),
                                        base_dir = self.parent.p.logs_dir,
                                        modules  = self.modules,
                                        **params)
        # Return the Job ID #
        return self.slurm_job.run()



# Built-in modules #
import time, inspect
import pickle

# Internal modules #
from autopaths.file_path import FilePath

# Third party modules #
from decorator import decorator

################################################################################
def cached(f):
    """Decorator for functions evaluated only once."""
    def memoized(*args, **kwargs):
        if hasattr(memoized, '__cache__'):
            return memoized.__cache__
        result = f(*args, **kwargs)
        memoized.__cache__ = result
        return result
    return memoized

###############################################################################
class property_cached(object):
    """
    New implementation of the property cached decorator.

    It converts a method with a single self argument
    into a property cached on the instance.

    In other words, it's like @property but memoized.

        from plumbing.cache import cached_property
        class Square(object):
            def __init__(self, size):
                self.size = size

            @cached_property
            def area(self):
                print("Evaluating...")
                return self.size * self.size

        shape = Square(5)
        print(shape.size)
        print(shape.area)
        shape.area = 99
        print(shape.area)
        del shape.area
        print(shape.area)
    """

    def __init__(self, func):
        self.func    = func
        self.__doc__ = getattr(func, '__doc__')
        self.name    = self.func.__name__

    def __get__(self, instance, owner):
        # If called from a class #
        if instance is None: return self
        # Does a cache exist for this instance? #
        self.check_cache(instance)
        # Is the answer in the cache? #
        if self.name in instance.__cache__: return instance.__cache__[self.name]
        # If not we will compute it #
        if inspect.isgeneratorfunction(self.func): result = tuple(self.func(instance))
        else:                                      result = self.func(instance)
        instance.__cache__[self.name] = result
        return result

    def __set__(self, instance, value):
        # Does a cache exist for this instance? #
        self.check_cache(instance)
        # Overwrite the value #
        instance.__cache__[self.name] = value

    def __delete__(self, instance):
        # Does a cache exist for this instance? #
        self.check_cache(instance)
        # Remove the key #
        instance.__cache__.pop(self.name, None)

    def check_cache(self, instance):
        if '__cache__' not in instance.__dict__: instance.__cache__ = {}

################################################################################
def property_pickled(f):
    """Same thing as above but the result will be stored on disk
    The path of the pickle file will be determined by looking for the
    `cache_dir` attribute of the instance containing the cached property.
    If no `cache_dir` attribute exists the `p` attribute will be accessed with
    the name of the property being cached."""
    # Called when you access the property #
    def retrieve_from_cache(self):
        # Is it in the cache ? #
        if '__cache__' not in self.__dict__: self.__cache__ = {}
        if f.__name__ in self.__cache__: return self.__cache__[f.__name__]
        # Where should we look in the file system ? #
        if 'cache_dir' in self.__dict__:
            path = FilePath(self.__dict__['cache_dir'] + f.func_name + '.pickle')
        else:
            path = getattr(self.p, f.func_name)
        # Is it on disk ? #
        if path.exists:
            with open(path) as handle: result = pickle.load(handle)
            self.__cache__[f.__name__] = result
            return result
        # Otherwise let's compute it #
        result = f(self)
        with open(path, 'w') as handle: pickle.dump(result, handle)
        self.__cache__[f.__name__] = result
        return result
    # Called when you set the property #
    def overwrite_cache(self, value):
        # Where should we look in the file system ? #
        if 'cache_dir' in self.__dict__:
            path = FilePath(self.__dict__['cache_dir'] + f.func_name + '.pickle')
        else:
            path = getattr(self.p, f.func_name)
        if value is None: path.remove()
        else: raise Exception("You can't set a pickled property, you can only delete it")
    # Return a wrapper #
    retrieve_from_cache.__doc__ = f.__doc__
    return property(retrieve_from_cache, overwrite_cache)

################################################################################
def expiry_every(seconds=0):
    def memoize_with_expiry(func, *args, **kwargs):
        # Get the cache #
        if not hasattr(func, '__cache__'): func.__cache__ = [(0,0)]
        cache = func.__cache__
        # Check the cache #
        if cache:
            result, timestamp = cache[0]
            age = time.time() - timestamp
            if age < seconds: return result
        # Update the cache #
        result = func(*args, **kwargs)
        cache[0] = (result, time.time())
        # Return #
        return result
    return decorator(memoize_with_expiry)

###############################################################################
class LazyString(object):
    """A string-like object that will only compute its value once, when accessed."""
    def __str__(self): return self.value

    def __init__(self, function):
        self._value = None
        self.function = function

    @property
    def value(self):
        if self._value is None: self._value = self.function()
        return self._value

###############################################################################
class LazyDict(object):
    """A dictionary-like object that will only compute its value once, when accessed."""
    def __getitem__(self, item): return self.value[item]

    def __init__(self, function):
        self._value = None
        self.function = function

    @property
    def value(self):
        if self._value is None: self._value = self.function()
        return self._value

###############################################################################
class LazyList(object):
    """A list-like object that will only compute its value once, when accessed."""
    def __iter__(self): return iter(self.value)
    def __len__(self):  return len(self.value)

    def __init__(self, function):
        self._value = None
        self.function = function

    @property
    def value(self):
        if self._value is None: self._value = self.function()
        return self._value

###############################################################################
class class_property(property):
    """You can use this like this:
        class Foo(object):
            @class_property
            @classmethod
            @cached
            def bar(cls):
                return 1+1
    """
    def __get__(self, cls, owner):
        return self.fget.__get__(None, owner)()

###############################################################################
class invalidate_cache(object):
    """
    Adds a hook to allow for another property setter to
    invalidated a part of the cache.

        class Square(object):
            def __init__(self, size):
                self._size = size

            @property
            def size(self):
                return self._size

            @cached_property
            def area(self):
                print("Evaluating...")
                return self.size * self.size

            @size.setter
            @invalidate_cache("area")
            def size(self, size):
                self._size = size

    shape = Square(5)
    print(shape.size)
    print(shape.area)
    shape.size = 6
    print(shape.area)
    """

    def __init__(self, func):
        self.func    = func
        self.__doc__ = getattr(func, '__doc__')